#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In this module you find the base workflow for a EOS calculation and
some helper methods to do so with AiiDA
"""
import copy
import numpy as np
from masci_tools.io.kkr_params import kkrparams
from masci_tools.io.common_functions import get_Ry2eV
from ase.eos import EquationOfState
from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiida.common.exceptions import NotExistentAttributeError
from aiida.engine import WorkChain, ToContext
from aiida.engine import calcfunction
from aiida_kkr.tools.common_workfunctions import update_params_wf
from aiida_kkr.workflows.voro_start import kkr_startpot_wc
from aiida_kkr.workflows.kkr_scf import kkr_scf_wc
from aiida_kkr.tools.save_output_nodes import create_out_dict_node

__copyright__ = (u'Copyright (c), 2018, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.9.2'
__contributors__ = u'Philipp Rüßmann'


class kkr_eos_wc(WorkChain):
    """
    Workchain of an equation of states calculation with KKR.

    Layout of the workflow:
      1. determine V0, scale_range, etc. from input
      2. run voro_start for V0 and smallest volume
          2.1 get minimum for RMTCORE (needs to be fixed for all calculations to be able to compare total energies
      3. submit kkr_scf calculations for all volumes using RMTCORE setting determined in step 2
      4. collect results
    """

    _workflowversion = __version__
    _wf_label = 'kkr_eos_wc_{}'  # replace with structure formula
    # replace with structure formula
    _wf_description = 'Equation of states workflow for {} using KKR'
    # workflow options (computer settings)
    _options_default = AttributeDict()
    # Queue name to submit jobs too
    _options_default.queue_name = ''
    _options_default.resources = AttributeDict()
    _options_default.resources.num_machines = 1
    # walltime in seconds after which the job gets killed (gets parsed to KKR)
    _options_default.max_wallclock_seconds = 60 * 60
    # execute KKR with mpi or without
    _options_default.withmpi = True
    # some additional scheduler commands
    # (e.g. project numbers in job scripts, OpenMP settings, ...)
    _options_default.custom_scheduler_commands = ''
    # workflow settings
    _wf_default = AttributeDict()
    # range around volume of starting structure which eos is computed
    _wf_default.scale_range = [0.94, 1.06]
    # number of calculations around
    _wf_default.nsteps = 7
    # create and return a structure which has the ground state volume
    # determined by the fit used
    _wf_default.ground_state_structure = True
    # use seekpath to get primitive structure after scaling to reduce
    # computational time
    _wf_default.use_primitive_structure = True
    # fitfunction used to determine ground state volume
    # (see ase.eos.EquationOfState class for details)
    _wf_default.fitfunction = 'birchmurnaghan'
    # settings for kkr_startpot behavior
    _wf_default.settings_kkr_startpot = kkr_startpot_wc.get_wf_defaults(silent=True)
    # settings for kkr_scf behavior
    _wf_default.settings_kkr_scf = kkr_scf_wc.get_wf_defaults(silent=True)[0]
    # change _wf_default of kkr_scf to deactivate DOS runs
    _wf_default.settings_kkr_scf['check_dos'] = False

    @classmethod
    def get_wf_defaults(cls, silent=False):
        """Print and return _wf_defaults dictionary.

        Can be used to easily create set of wf_parameters.
        returns _wf_defaults, _options_default
        """
        if not silent:
            print(f'Version of workflow: {cls._workflowversion}')
        return cls._wf_default, cls._options_default

    @classmethod
    def define(cls, spec):
        """
        Defines the outline of the workflow.
        """
        # Take input of the workflow or use defaults defined above
        super(kkr_eos_wc, cls).define(spec)
        spec.input(
            'options',
            valid_type=orm.Dict,
            required=False,  # computer options
            default=lambda: orm.Dict(dict=cls._options_default)
        )
        spec.input(
            'wf_parameters',
            valid_type=orm.Dict,
            required=False,  # workfunction settings
            default=lambda: orm.Dict(dict=cls._wf_default),
            help="""
            Workfunction specific parameters, controlling the behavior of the
            EOS workchain.
            """
        )
        spec.input(
            'kkr',
            valid_type=orm.Code,
            required=True,
            help="""
            Code entry for the KKRhost calculations.
            """
        )
        spec.input(
            'voronoi',
            valid_type=orm.Code,
            required=True,
            help="""
            Code entry for the Voronoi calculations.
            """
        )
        spec.input(
            'structure',
            valid_type=orm.StructureData,
            required=True,
            help="""
            Initial structure for which the EOS will be calculated.
            """
        )
        spec.input(
            'calc_parameters',
            valid_type=orm.Dict,
            required=False,
            help="""
            KKR input parameters. Everything (except structural factors) which
            would normally be written in the inputcard.
            """
        )

        # define output nodes
        spec.output(
            'eos_results',
            valid_type=orm.Dict,
            required=True,
        )
        spec.output(
            'gs_structure',
            valid_type=orm.StructureData,
            required=False,
        )
        spec.output(
            'explicit_kpoints',
            valid_type=orm.KpointsData,
            required=False,
        )
        spec.output(
            'get_explicit_kpoints_path_parameters',
            valid_type=orm.Dict,
            required=False,
        )

        # Here the structure of the workflow is defined
        spec.outline(
            # 1. initialize workflow and check input consistency
            cls.start,
            # 2. prepare structures
            cls.prepare_strucs,
            # 3. run voronoi calculation for smallest volume0
            cls.run_vorostart,
            # 4. check voronoi output and extract RMTCORE parameter
            cls.check_voro_out,
            # 5. submit KKR calculations for all steps
            cls.run_kkr_steps,
            # 6. collect output and fit results
            cls.collect_data_and_fit,
            # 7. collect and return output nodes
            cls.return_results
        )

        # ToDo: improve error codes
        spec.exit_code(
            221,
            'ERROR_INVALID_INPUT',
            message='ERROR: inputs invalid',
        )
        spec.exit_code(
            222,
            'ERROR_NOT_ENOUGH_SUCCESSFUL_CALCS',
            message='ERROR: need at least 3 successful calculations',
        )
        spec.exit_code(
            223,
            'ERROR_NSTEPS_TOO_SMALL',
            message='ERROR: nsteps is smaller than 3, need at least three data points to do fitting',
        )
        spec.exit_code(
            224,
            'ERROR_INVALID_FITFUN',
            message='given fitfunction name not valid',
        )
        spec.exit_code(
            225,
            'ERROR_VOROSTART_NOT_SUCCESSFUL',
            message='ERROR: kkr_startpot was not successful. Check you inputs.',
        )

    def start(self):
        """
        initialize context and check input nodes
        """

        self.report(f'INFO: starting KKR eos workflow version {self._workflowversion}')

        # now extract information from input nodes
        try:
            self.ctx.wf_options = self.inputs.get('options').get_dict()
            self.ctx.wf_parameters = self.inputs.get('wf_parameters').get_dict()
            # TODO: check if code is KKR code
            self.ctx.kkr = self.inputs.get('kkr')
            # TODO: check if code is voronoi code
            self.ctx.voro = self.inputs.get('voronoi')
            self.ctx.structure = self.inputs.get('structure')
            # optional, TODO: needs to be filled with defaults if not present
            self.ctx.calc_parameters = self.inputs.get('calc_parameters')
        except AttributeError:
            # in case of failure, exit workflow here
            return self.exit_codes.ERROR_INVALID_INPUT  # pylint: disable=no-member

        # add label and description if not given (contains structure name)
        # if self.label is None:
        self.ctx.label = self._wf_label.format(self.ctx.structure.get_formula())
        # if self.description is None:
        self.ctx.description = self._wf_description.format(self.ctx.structure.get_formula())
        if self.ctx.wf_parameters['settings_kkr_startpot'].get('_label', None) is None:
            self.ctx.wf_parameters['settings_kkr_startpot']['_label'] = (
                self.ctx.label + f'_kkr_startpot_{self.ctx.structure.get_formula()}'
            )
        if self.ctx.wf_parameters['settings_kkr_startpot'].get('_description', None) is None:
            self.ctx.wf_parameters['settings_kkr_startpot']['_description'] = (
                self.ctx.description + f' kkr_startpot step for {self.ctx.structure.get_formula()}'
            )
        if self.ctx.wf_parameters['settings_kkr_scf'].get('label', None) is None:
            self.ctx.wf_parameters['settings_kkr_scf']['label'] = (
                self.ctx.label + f'_kkr_scf_{self.ctx.structure.get_formula()}'
            )
        if self.ctx.wf_parameters['settings_kkr_scf'].get('description', None) is None:
            self.ctx.wf_parameters['settings_kkr_scf']['description'] = (
                self.ctx.description + f' kkr_scf step for {self.ctx.structure.get_formula()}'
            )

        # initialize some other things used to collect results etc.
        self.ctx.successful = True
        self.ctx.warnings = []
        self.ctx.rms_threshold = self.ctx.wf_parameters['settings_kkr_scf'].get(
            'convergence_criterion',
            10**-7,
        )
        self.ctx.nsteps = self.ctx.wf_parameters.get(
            'nsteps',
            self._wf_default.nsteps,
        )
        self.ctx.scale_range = self.ctx.wf_parameters.get(
            'scale_range',
            self._wf_default.scale_range,
        )
        # fitfunction used to get ground state structure
        self.ctx.fitfunc_gs_out = self.ctx.wf_parameters.get(
            'fitfunction',
            self._wf_default.fitfunction,
        )
        # boolean, return output structure or not
        self.ctx.return_gs_struc = self.ctx.wf_parameters.get(
            'ground_state_structure',
            self._wf_default.ground_state_structure,
        )
        self.ctx.use_primitive_structure = self.ctx.wf_parameters.get(
            'use_primitive_structure',
            self._wf_default.use_primitive_structure,
        )
        self.ctx.scaled_structures = []  # filled in prepare_strucs
        self.ctx.fitnames = [
            'sj',
            'taylor',
            'murnaghan',
            'birch',
            'birchmurnaghan',
            'pouriertarantola',
            'vinet',
            'antonschmidt',
            'p3',
        ]  # list of allowed fits
        self.ctx.sub_wf_ids = {}  # filled with workflow uuids

        # check input
        if self.ctx.nsteps < 3:
            return self.exit_codes.ERROR_NSTEPS_TOO_SMALL  # pylint: disable=no-member
        if self.ctx.fitfunc_gs_out not in self.ctx.fitnames:
            return self.exit_codes.ERROR_INVALID_FITFUN  # pylint: disable=no-member

        # set scale_factors from scale_range and nsteps
        self.ctx.scale_factors = np.linspace(
            start=self.ctx.scale_range[0],
            stop=self.ctx.scale_range[1],
            num=self.ctx.nsteps,
        )
        return None

    def prepare_strucs(self):
        """Create new set of scaled structures for the E-V curve.

        The structures are generated using the 'rescale' workfunction
        (see end of the workflow)
        """
        for scale_fac in self.ctx.scale_factors:
            scaled_structure = rescale(self.ctx.structure, orm.Float(scale_fac))
            if self.ctx.use_primitive_structure:
                scaled_structure = get_primitive_structure(
                    scaled_structure,
                    orm.Bool(False),
                )
            self.ctx.scaled_structures.append(scaled_structure)

    def run_vorostart(self):
        """Run vorostart workflow for smallest structure to determine rmtcore.

        One needs to run all the calculations with the same rmtcore to be able
        to compare the energies. This value is then set for all others.
        """
        wfd = kkr_startpot_wc.get_wf_defaults(silent=True)
        set_keys = []
        # first set options
        for key in list(self.ctx.wf_options.keys()):
            wfd[key] = self.ctx.wf_options.get(key)
            set_keys.append(key)
        # then set ef_settings
        vorostart_settings = self.ctx.wf_parameters.get('settings_kkr_startpot')
        for key in list(vorostart_settings.keys()):
            # skip setting of options (done above already)
            if key not in set_keys:
                wfd[key] = vorostart_settings[key]
        scaled_struc = self.ctx.scaled_structures[0]
        future = self.submit(
            kkr_startpot_wc,
            structure=scaled_struc,
            kkr=self.ctx.kkr,
            voronoi=self.ctx.voro,
            wf_parameters=orm.Dict(dict=wfd),
            calc_parameters=self.ctx.calc_parameters,
            options=orm.Dict(dict=self.ctx.wf_options)
        )

        self.report(f'INFO: running kkr_startpot workflow (pk= {future.pk})')
        self.ctx.sub_wf_ids['kkr_startpot_1'] = future.uuid

        return ToContext(kkr_startpot=future)

    def check_voro_out(self):
        """Check output of the vorostart workflow.

        The outputs are then used to create inputs for the next set of
        calculations (rmtcore setting etc.)
        """
        self.report('INFO: checking voronoi output')
        # get output of kkr_startpot
        out_wc = self.ctx.kkr_startpot
        res = out_wc.outputs.results_vorostart_wc
        voro_params = out_wc.outputs.last_params_voronoi
        smallest_voro_remote = out_wc.outputs.last_voronoi_remote
        smallest_voro_results = out_wc.outputs.last_voronoi_results
        vorostart_success = res.get_dict().get('successful', False)

        if vorostart_success:
            rmt = []
            radii = smallest_voro_results.get_dict()['radii_atoms_group']
            for rad_iatom in radii:
                if 'rmt0' in list(rad_iatom.keys()):
                    rmt.append(rad_iatom['rmt0'])
            # needs to be multiplied by alat in atomic units!
            rmtcore_min = np.array(rmt) * smallest_voro_results.get_dict().get('alat')
            self.report(f'INFO: extracted rmtcore_min ({rmtcore_min})')
        else:
            self.report(f'ERROR: kkr_startpot workflow not successful')
            return self.exit_codes.ERROR_VOROSTART_NOT_SUCCESSFUL  # pylint: disable=no-member

        # update parameter node with rmtcore setting
        voro_params_with_rmtcore = kkrparams(**voro_params.get_dict())
        voro_params_with_rmtcore.set_value('<RMTCORE>', rmtcore_min)
        voro_params_with_rmtcore_dict = voro_params_with_rmtcore.get_dict()
        voro_params_with_rmtcore = update_params_wf(
            voro_params,
            orm.Dict(dict=voro_params_with_rmtcore_dict),
        )
        self.report(f'INFO: updated kkr_parameters including RMTCORE setting (uuid={voro_params_with_rmtcore.uuid})')

        # store links to context
        self.ctx.params_kkr_run = voro_params_with_rmtcore
        self.ctx.smallest_voro_remote = smallest_voro_remote

        return None

    def run_kkr_steps(self):
        """Submit KKR calculations for all structures.

        This will skip the vorostart step for smallest structure.
        """

        self.report('INFO: running kkr scf steps')
        # params for scf wfd
        wfd = kkr_scf_wc.get_wf_defaults(silent=True)[0]
        set_keys = []
        # first set options
        for key in list(self.ctx.wf_options.keys()):
            wfd[key] = self.ctx.wf_options.get(key)
            set_keys.append(key)
        # then set ef_settings
        kkr_scf_settings = self.ctx.wf_parameters.get('settings_kkr_scf')
        for key in list(kkr_scf_settings.keys()):
            # skip setting of options (done above already)
            if key not in set_keys:
                wfd[key] = kkr_scf_settings[key]

        # used to collect all submitted calculations
        calcs = {}

        # submit first calculation separately
        self.report(
            f'submit calc for scale fac= {self.ctx.scale_factors[0]} on {self.ctx.scaled_structures[0].get_formula()}'
        )
        future = self.submit(
            kkr_scf_wc,
            kkr=self.ctx.kkr,
            remote_data=self.ctx.smallest_voro_remote,
            wf_parameters=orm.Dict(dict=wfd),
            calc_parameters=self.ctx.params_kkr_run,
            options=orm.Dict(dict=self.ctx.wf_options)
        )
        scale_fac = self.ctx.scale_factors[0]
        calcs[f'kkr_1_{scale_fac}'] = future
        self.ctx.sub_wf_ids['kkr_scf_1'] = future.uuid

        # then also submit the rest of the calculations
        for i in range(len(self.ctx.scale_factors) - 1):
            scale_fac = self.ctx.scale_factors[i + 1]
            scaled_struc = self.ctx.scaled_structures[i + 1]
            self.report(f'submit calc for scale fac= {scale_fac} on {scaled_struc.get_formula()}')
            future = self.submit(
                kkr_scf_wc,
                structure=scaled_struc,
                kkr=self.ctx.kkr,
                voronoi=self.ctx.voro,
                wf_parameters=orm.Dict(dict=wfd),
                calc_parameters=self.ctx.params_kkr_run,
                options=orm.Dict(dict=self.ctx.wf_options)
            )
            calcs[f'kkr_{i+2}_{scale_fac}'] = future
            self.ctx.sub_wf_ids[f'kkr_scf_{i+2}'] = future.uuid

        # save uuids of calculations to context
        self.ctx.kkr_calc_uuids = []
        # sorting important to have correct assignment of scaling and structure
        # info later on
        for name in np.sort(list(calcs.keys())):
            calc = calcs[name]
            self.ctx.kkr_calc_uuids.append(calc.uuid)

        self.report(f'INFO: submitted calculations: {calcs}')

        return ToContext(**calcs)

    def collect_data_and_fit(self):
        """Collect output of KKR calculations and perform eos fitting."""
        self.report('INFO: collect kkr results and fit data')
        calc_uuids = self.ctx.kkr_calc_uuids
        etot = []
        for iic, uuid in enumerate(calc_uuids):
            node = orm.load_node(uuid)
            try:
                d_result = node.outputs.output_kkr_scf_wc_ParameterResults.get_dict()
                self.report(
                    f'INFO: extracting output of calculation {uuid}: successful={d_result[u"successful"]}, rms={d_result[u"convergence_value"]}'
                )
                if d_result[u'successful']:
                    pk_last_calc = d_result['last_calc_nodeinfo']['pk']
                    node_2 = orm.load_node(pk_last_calc)
                    scale = self.ctx.scale_factors[iic]
                    ener = node_2.outputs.output_parameters.get_dict()['total_energy_Ry']
                    rms = d_result[u'convergence_value']
                    scaled_struc = self.ctx.scaled_structures[iic]
                    vol = scaled_struc.get_cell_volume()
                    alat = scaled_struc.cell_lengths[0]
                    # Only take those calculations which
                    if rms <= self.ctx.rms_threshold:
                        etot.append([scale, ener, vol, rms, alat])
                    else:
                        warn = f'rms of calculation with uuid={uuid} not low enough ({rms} > {self.ctx.rms_threshold})'

                        self.report(f'WARNING: {warn}')
                        self.ctx.warnings.append(warn)
            except (AttributeError, NotExistentAttributeError):
                warn = f'calculation with uuid={uuid} not successful'
                self.report(f'WARNING: {warn}')
                self.ctx.warnings.append(warn)

        # collect calculation outcome
        etot = np.array(etot)
        self.report(f'INFO: collected data from calculations= {etot}')

        # check if at least 3 points were successful (otherwise fit does not work)
        if len(etot) < 3:
            return self.exit_codes.ERROR_NOT_ENOUGH_SUCCESSFUL_CALCS  # pylint: disable=no-member

        scalings = etot[:, 0]
        rms = etot[:, -1]
        # convert to eV and per atom units
        etot = etot / len(scaled_struc.sites)  # per atom values
        etot[:, 1] = etot[:, 1] * get_Ry2eV()  # convert energy from Ry to eV
        volumes, energies = etot[:, 2], etot[:, 1]

        # do multiple fits to data
        self.report('INFO: output of fits:')
        self.report(f'{"fitfunc":18} {"v0":8} {"e0":7} {"B":7} {"alat":7}')
        self.report('-----------------------------------------')
        fitnames = self.ctx.fitnames
        alldat = []
        fitdata = {}
        for fitfunc in fitnames:
            try:
                eos = EquationOfState(volumes, energies, eos=fitfunc)
                v0, e0, B = eos.fit()
                py_func = scaled_struc.get_pymatgen()
                py_func.scale_lattice(v0)
                alat = py_func.lattice.a
                fitdata[fitfunc] = [v0, e0, B, alat]
                alldat.append([v0, e0, B, alat])
                self.report(f'{fitfunc:16} {v0:8.3f} {e0:7.3f} {B:7.3f} {alat:7.3f}')
            # capture all errors and mark fit as unsuccessful
            except (ValueError, RuntimeError, UserWarning, RuntimeWarning):
                self.ctx.warnings.append(f'fit unsuccessful for {fitfunc} function')
                if fitfunc == self.ctx.fitfunc_gs_out:
                    self.ctx.successful = False
        alldat = np.array(alldat)
        self.report('-----------------------------------------')
        self.report(
            f'{"mean":16} {np.mean(alldat[:, 0]):8.3f} {np.mean(alldat[:, 1]):7.3f} {np.mean(alldat[:, 2]):7.3f} {np.mean(alldat[:, 3]):7.3f}'
        )
        self.report(
            f'{"std":16} {np.std(alldat[:, 0]):8.3f} {np.std(alldat[:, 1]):7.3f} {np.std(alldat[:, 2]):7.3f} {np.std(alldat[:, 3]):7.3f}'
        )

        # store results in context
        self.ctx.volumes = volumes
        self.ctx.energies = energies
        self.ctx.scalings = scalings
        self.ctx.rms = rms
        self.ctx.fitdata = fitdata
        self.ctx.fit_mean_values = {
            '<v0>': np.mean(alldat[:, 0]),
            '<e0>': np.mean(alldat[:, 1]),
            '<B>': np.mean(alldat[:, 2]),
            '<alat>': np.mean(alldat[:, 3])
        }
        self.ctx.fit_std_values = {
            's_v0': np.std(alldat[:, 0]),
            's_e0': np.std(alldat[:, 1]),
            's_B': np.std(alldat[:, 2]),
            's_alat': np.std(alldat[:, 3])
        }

        return None

    def return_results(self):
        """Create output dictionary and run output node generation."""
        self.report('INFO: create output node')
        outdict = {}
        outdict['successful'] = self.ctx.successful
        outdict['warnings'] = self.ctx.warnings
        outdict['sub_workflow_uuids'] = self.ctx.sub_wf_ids
        outdict['nsteps_input'] = self.ctx.nsteps
        outdict['scale_range_input'] = self.ctx.scale_range
        outdict['scale_factors_all'] = self.ctx.scale_factors
        outdict['volumes'] = self.ctx.volumes
        outdict['energies'] = self.ctx.energies
        outdict['scalings'] = self.ctx.scalings
        outdict['rms'] = self.ctx.rms
        outdict['parameter_fits'] = self.ctx.fitdata
        outdict['fits_mean'] = self.ctx.fit_mean_values
        outdict['fits_std'] = self.ctx.fit_std_values
        outdict['formula'] = self.ctx.structure.get_formula()
        outdict['label'] = self.ctx.label
        if self.ctx.successful and self.ctx.return_gs_struc:
            # final result: scaling factor for equilibrium
            v0, e0, B, alat = self.ctx.fitdata.get(self.ctx.fitfunc_gs_out)
            scale_fac0 = v0 / self.ctx.structure.get_cell_volume() * len(self.ctx.structure.sites)
            outdict['gs_scale_factor'] = scale_fac0
            outdict['gs_fitfunction'] = self.ctx.fitfunc_gs_out
            gs_structure = rescale(self.ctx.structure, orm.Float(scale_fac0))
            if self.ctx.use_primitive_structure:
                tmpdict = get_primitive_structure(gs_structure, orm.Bool(True))
                conv_structure, explicit_kpoints, parameters, gs_structure = tmpdict['conv_structure'], tmpdict[
                    'explicit_kpoints'], tmpdict['parameters'], tmpdict['primitive_structure']
                outdict['gs_kpoints_seekpath_params_uuid'] = parameters.uuid
            gs_structure.label = (f'ground_state_structure_{gs_structure.get_formula()}')
            gs_structure.description = (
                f'Ground state structure of {gs_structure.get_formula()} after running eos workflow. Uses {self.ctx.fitfunc_gs_out} fit.'
            )
            outdict['gs_structure_uuid'] = gs_structure.uuid

        # create output nodes in dict with link names
        outnodes = {}
        if self.ctx.successful and self.ctx.return_gs_struc:
            outnodes['gs_structure'] = gs_structure
            if self.ctx.use_primitive_structure:
                outnodes['explicit_kpoints'] = explicit_kpoints
                outnodes['get_explicit_kpoints_path_parameters'] = parameters

        # create results node with calcfunction for data provenance
        link_nodes = outnodes.copy()
        for wf_label, sub_wf_uuid in self.ctx.sub_wf_ids.items():
            if 'kkr_scf' in wf_label:
                link_nodes[wf_label] = orm.load_node(sub_wf_uuid).outputs.output_kkr_scf_wc_ParameterResults
            else:
                link_nodes[wf_label] = orm.load_node(sub_wf_uuid).outputs.results_vorostart_wc
        outnodes['eos_results'] = create_out_dict_node(orm.Dict(dict=outdict), **link_nodes)

        # set out nodes and corresponding link names
        for link_name, node in outnodes.items():
            self.out(link_name, node)


### Helper functions and workfunctions ###


def rescale_no_wf(structure, scale) -> orm.StructureData:
    """
    Rescales a crystal structure. DOES NOT keep the provenance in the database.

    :param structure, a StructureData node (pk, or uuid)
    :param scale, float scaling factor for the cell

    :returns: New StructureData node with rescalled structure, which is linked to input Structure
              and None if inp_structure was not a StructureData

    copied and modified from aiida_fleur.tools.StructureData_util
    """

    scaled_volume = structure.get_cell_volume() * scale.value
    scaled_structure = copy.deepcopy(structure.get_pymatgen())
    scaled_structure.scale_lattice(scaled_volume)

    return orm.StructureData(pymatgen=scaled_structure)


@calcfunction
def rescale(inp_structure, scale):
    """
    Rescales a crystal structure. Keeps the provenance in the database.

    :param inp_structure, a StructureData node (pk, or uuid)
    :param scale, float scaling factor for the cell

    :returns: New StructureData node with rescalled structure, which is linked to input Structure
              and None if inp_structure was not a StructureData

    copied and modified from aiida_fleur.tools.StructureData_util
    """

    return rescale_no_wf(inp_structure, scale)


@calcfunction
def get_primitive_structure(structure, return_all):
    """
    calls get_explicit_kpoints_path which gives primitive structure
    auxiliary workfunction to keep provenance
    """
    from aiida.tools import get_explicit_kpoints_path

    output = get_explicit_kpoints_path(structure)
    conv_structure = output['conv_structure']
    explicit_kpoints = output['explicit_kpoints']
    parameters = output['parameters']
    primitive_structure = output['primitive_structure']

    if return_all:
        return {
            'conv_structure': conv_structure,
            'explicit_kpoints': explicit_kpoints,
            'parameters': parameters,
            'primitive_structure': primitive_structure
        }

    return primitive_structure



# -*- coding: utf-8 -*-
"""
This contains code snippets and utility useful for dealing with parameter data nodes
commonly used by the plugin and workflows
"""


def clean_nones(dict_to_clean):
    """Recursively remove all keys which values are None from a nested dictionary
    return the cleaned dictionary

    :param dict_to_clean: (dict): python dictionary to remove keys with None as value
    :return: dict, cleaned dictionary
    """
    new_dict = {}
    for key, val in dict_to_clean.items():
        if isinstance(val, dict):
            new_val = clean_nones(val)
        else:
            new_val = val
        if new_val is not None:  # currently we keep empty dicts
            new_dict[key] = new_val

    return new_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tools for the impurity caluclation plugin and its workflows
"""

from builtins import object, str
from masci_tools.io.common_functions import open_general
from masci_tools.io.common_functions import search_string
import numpy as np
from masci_tools.io.common_functions import get_alat_from_bravais
from masci_tools.io.common_functions import vec_to_angles
from aiida.common.constants import elements as PeriodicTableElements
from aiida.common.exceptions import InputValidationError

__copyright__ = (u'Copyright (c), 2018, Forschungszentrum Jülich GmbH,'
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.7.0'
__contributors__ = u'Philipp Rüßmann'


class modify_potential(object):
    """
    Class for old modify potential script, ported from modify_potential script, initially by D. Bauer
    """

    def _check_potstart(self, str1, mode='pot', shape_ver='new'):
        if mode == 'shape':
            if shape_ver == 'new':
                check1 = 'Shape number' in str1
            else:
                check1 = (len(str1) == 11)
        else:
            check1 = 'exc:' in str1
        return check1

    def _read_input(self, filepath):
        with open_general(filepath) as f:
            data = f.readlines()

        # check if mode is shapefun or potential
        mode = 'shape'
        for line in data:
            if 'exc:' in line:
                mode = 'pot'
                break

        # read file
        index1 = []
        index2 = []
        for i in range(len(data)):
            if self._check_potstart(data[i], mode=mode):
                index1.append(i)
                if len(index1) > 1:
                    index2.append(i - 1)
        index2.append(i)

        # read shapefun if old style is used
        if mode == 'shape' and len(index1) < 1:
            index1 = []
            index2 = []
            for i in range(len(data)):
                if self._check_potstart(data[i], mode=mode, shape_ver='old'):
                    index1.append(i)
                if len(index1) > 1:
                    index2.append(i - 1)
                index2.append(i)
        """
        print(index1)
        print(index2)

        print('Potential file read')
        print('found %i potentials in file'%len(index1))
        print('')
        """

        return index1, index2, data

    def shapefun_from_scoef(self, scoefpath, shapefun_path, atom2shapes, shapefun_new):
        """
        Read shapefun and create impurity shapefun using scoef info and shapes array

        :param scoefpath: absolute path to scoef file
        :param shapefun_path: absolute path to input shapefun file
        :param shapes: shapes array for mapping between atom index and shapefunction index
        :param shapefun_new: absolute path to output shapefun file to which the new shapefunction will be written
        """
        index1, index2, data = self._read_input(shapefun_path)

        order = list(range(len(index1)))

        natomtemp = int(open_general(scoefpath).readlines()[0])
        with open_general(scoefpath) as f:
            filedata = f.readlines()[1:natomtemp + 1]
        listnew = []
        for line in filedata:
            if (len(line.split()) > 1):
                listnew.append(atom2shapes[int(line.split()[3]) - 1] - 1)
        order = listnew

        datanew = []
        for i in range(len(order)):
            for ii in range(index1[order[i]], index2[order[i]] + 1):
                datanew.append(data[ii])

        # add header to shapefun_new
        tmp = datanew
        datanew = []
        datanew.append(f'   {len(order)}\n')
        datanew.append('  1.000000000000E+00\n')
        datanew += tmp
        with open_general(shapefun_new, 'w') as f:
            f.writelines(datanew)

    def neworder_potential(
        self, potfile_in, potfile_out, neworder, potfile_2=None, replace_from_pot2=None, debug=False
    ):
        """
        Read potential file and new potential using a list describing the order of the new potential.
        If a second potential is given as input together with an index list, then the corresponding of
        the output potential are overwritten with positions from the second input potential.

        :param potfile_in: absolute path to input potential
        :type potfile_in: str
        :param potfile_out: absolute path to output potential
        :type potfile_out: str
        :param neworder: list after which output potential is constructed from input potential
        :type neworder: list
        :param potfile_2: optional, absolute path to second potential file if
            positions in new list of potentials shall be replaced by positions of
            second potential, requires *replace_from_pot* to be given as well
        :type potfile_2: str
        :param replace_from_pot: optional, list containing tuples of (position
            in newlist that is to be replaced, position in pot2 with which position
            is replaced)
        :type replace_from_pot: list

        :usage:
            1. modify_potential().neworder_potential(<path_to_input_pot>, <path_to_output_pot>, [])
        """

        # read potfile
        if debug:
            print(f'potfile: {potfile_in}')
        index1, index2, data = self._read_input(potfile_in)
        if debug:
            print(index1, index2)
            # add label to potname
            ii = 0
            for i in range(len(data)):
                line = data[i]
                if 'exc:' in line:
                    line = line.replace('     exc:', '%.4i exc:' % ii)
                    data[i] = line
                    ii += 1

        # read potfile2
        if potfile_2 is not None:
            if debug:
                print(f'potfile2: {potfile_2}')
            index12, index22, data2 = self._read_input(potfile_2)
            if debug:
                print(index12, index22)
                # add label to potname
                ii = 0
                for i in range(len(data2)):
                    line = data2[i]
                    if 'exc:' in line:
                        line = line.replace('     exc:', '%.4i exc:' % ii)
                        data2[i] = line
                        ii += 1
            # check if also replace_from_pot2 is given correctly
            if replace_from_pot2 is None:
                raise ValueError('replace_from_pot2 not given')
            else:
                replace_from_pot2 = np.array(replace_from_pot2)
                if debug:
                    print('replace list:', replace_from_pot2)
                if np.shape(replace_from_pot2)[1] != 2:
                    raise ValueError('replace_from_pot2 needs to be a 2D array!')
        else:
            if replace_from_pot2 is not None:
                raise ValueError('replace_from_pot2 given but potfile_2 not given')

        # set order in which potential file is written
        # ensure that numbers are integers:
        order = [int(i) for i in neworder]
        if debug:
            print('neworder:', order)

        # create output potential
        datanew = []
        for i in range(len(order)):
            # check if new position is replaced with position from old pot
            if replace_from_pot2 is not None and i in replace_from_pot2[:, 0]:
                replace_index = replace_from_pot2[replace_from_pot2[:, 0] == i][0][1]
                if debug:
                    print(i, replace_index)
                for count, ii in enumerate(range(index12[replace_index], index22[replace_index] + 1)):
                    if count == 0 and debug:
                        print(data2[ii])
                    datanew.append(data2[ii])
            else:  # otherwise take new potntial according to input list
                if debug:
                    print(i)
                for count, ii in enumerate(range(index1[order[i]], index2[order[i]] + 1)):
                    if count == 0 and debug:
                        print(data[ii])
                    datanew.append(data[ii])

        # write out new potential
        with open_general(potfile_out, 'w') as f:
            f.writelines(datanew)


####################################################################################
# for create scoef functions


def get_structure_data(structure):
    """
    Function to take data from AiiDA's StructureData type and store it into a single numpy array of the following form:
    a = [[x-Position 1st atom, y-Position 1st atom, z-Position 1st atom, index 1st atom, charge 1st atom, 0.],
         [x-Position 2nd atom, y-Position 2nd atom, z-Position 2nd atom, index 2nd atom, charge 1st atom, 0.],
         [..., ..., ..., ..., ..., ...],
         ...
         ]

    :param structure: input structure of the type StructureData

    :return: numpy array a[# of atoms in the unit cell][5] containing the structure related data (positions in units
             of the unit cell length)

    :note:
    """

    #get the connection between coordination number and element symbol
    _atomic_numbers = {data['symbol']: num for num, data in PeriodicTableElements.items()}

    #initialize the array that will be returned later (it will be a (# of atoms in the cell) x 6-matrix)
    a = np.zeros((len(structure.sites), 6))
    k = 0  #running index for filling up the array with data correctly
    charges = []  #will be needed to return the charge number for the different atoms later

    #loop to fill up the a-array with positions, index, charge and a 0. for every atom in the cell
    sites = structure.sites
    n = len(structure.sites) + 1  #needed to do the indexing of atoms
    m = len(structure.sites)  #needed to do the indexing of atoms
    for site in sites:
        for j in range(3):
            a[k][j] = site.position[j]
        sitekind = structure.get_kind(site.kind_name)
        naez = n - m
        m = m - 1
        #convert the element symbol from StructureData to the charge number
        if sitekind.get_symbols_string() == '{H0.00X1.00}':
            # special case when HX is mistaken for H instead of X
            charges.append(0.0)
        else:
            for ikind in range(len(sitekind.symbols)):
                site_symbol = sitekind.symbols[ikind]
                charges.append(_atomic_numbers[site_symbol])
        i = len(charges) - 1
        a[k][3] = int(naez)
        a[k][4] = float(charges[i])
        k = k + 1

    return a


def select_reference(structure_array, i):
    """
    Function that references all of the atoms in the cell to one particular atom i in the cell and calculates
    the distance from the different atoms to atom i. New numpy array will have the form:
    x = [[x-Position 1st atom, y-Position 1st atom, z-Position 1st atom, index 1st atom, charge 1st atom,
            distance 1st atom to atom i],
         [x-Position 2nd atom, y-Position 2nd atom, z-Position 2nd atom, index 2nd atom, charge 1st atom,
            distance 1st atom to atom i],
         [..., ..., ..., ..., ..., ...],
         ...
         ]

    :param structure_array: input array of the cell containing all the atoms (obtained from get_structure_data)
    :param i: index of the atom which should be the new reference

    :return: new structure array with the origin at the selected atom i (for KKRimp: impurity atom)

    :note: the first atom in the structure_array is labelled with 0, the second with 1, ...
    """

    #initialize new array for data centered around atom i
    x = np.zeros((len(structure_array), 6))

    #take the positions of atom i as new origin
    x_ref = np.array([structure_array[i][0], structure_array[i][1], structure_array[i][2], 0, 0, 0])

    #calculate the positions and distances for all atoms in the cell with respect to the chosen atom i
    for k in range(len(structure_array)):
        x[k][5] = get_distance(structure_array, i, k)
        for j in range(5):
            x[k][j] = structure_array[k][j] - x_ref[j]

    return x


def get_distance(structure_array, i, j):
    """
    Calculates and returns the distances between to atoms i and j in the given structure_array

    :param structure_array: input numpy array of the cell containing all the atoms ((# of atoms) x 6-matrix)
    :params i, j: indices of the atoms for which the distance should be calculated (indices again start at 0)

    :return: distance between atoms i and j in units of alat

    :note:
    """

    #calculate x-, y-, z-components of distance of i and j
    del_x = structure_array[i][0] - structure_array[j][0]
    del_y = structure_array[i][1] - structure_array[j][1]
    del_z = structure_array[i][2] - structure_array[j][2]

    #return absolute value of the distance of atom i and j
    return np.sqrt(del_x * del_x + del_y * del_y + del_z * del_z)


def rotate_onto_z(structure, structure_array, vector):
    """
    Rotates all positions of a structure array of orientation 'orient' onto the z-axis. Needed to implement the
    cylindrical cutoff shape.

    :param structure: input structure of the type StructureData
    :param structure_array: input structure array, obtained by select_reference for the referenced system.
    :param vector: reference vector that has to be mapped onto the z-axis.

    :return: rotated system, now the 'orient'-axis is aligned with the z-axis
    """

    #get angles, from vector
    angles = vec_to_angles(vector)
    theta = np.array(angles[1])
    phi = np.array(angles[2])

    #initialize needed arrays
    x_res = np.delete(structure_array, np.s_[3:6], 1)
    x_temp_1 = np.delete(structure_array, np.s_[3:6], 1)
    x_temp_2 = np.delete(structure_array, np.s_[3:6], 1)

    #define rotation matrices
    #========================
    #rotation around z-axis with angle phi
    R_z = np.array([[np.cos(-phi), -np.sin(-phi), 0.], [np.sin(-phi), np.cos(-phi), 0.], [0., 0., 1]])
    #rotation around y-axis with angle theta
    R_y = np.array([[np.cos(-theta), 0, np.sin(-theta)], [0., 1., 0.], [-np.sin(-theta), 0., np.cos(-theta)]])

    #first rotate around z-axis
    for i in range(len(structure_array)):
        x_temp_1[i] = np.dot(R_z, x_res[i])
        x_temp_2[i] = np.dot(R_y, x_temp_1[i])

    return x_temp_2


def find_neighbors(structure, structure_array, i, radius, clust_shape='spherical', h=0., vector=[0., 0., 1.]):
    """
    Applies periodic boundary conditions and obtains the distances between the selected atom i in the cell and
    all other atoms that lie within a cutoff radius r_cut. Afterwards an numpy array with all those atoms including
    atom i (x_res) will be returned.

    :param structure: input parameter of the StructureData type containing the three bravais lattice cell vectors
    :param structure_array: input numpy structure array containing all the structure related data
    :param i: centered atom at which the origin lies (same one as in select_reference)
    :param radius: Specifies the radius of the cylinder or of the sphere, depending on clust_shape.
                   Input in units of the lattice constant.
    :param clust_shape: specifies the shape of the cluster that is used to determine the neighbors for the 'scoef' file.
                        Default value is 'spherical'. Other possible forms are 'cylindrical' ('h' and 'orient'
                        needed), ... .
    :param h: needed for a cylindrical cluster shape. Specifies the height of the cylinder. Default=0.
              Input in units of the lattice constant.
    :param vector: needed for a cylindrical cluster shape. Specifies the orientation vector of the cylinder. Default:
                   z-direction.

    :return: array with all the atoms within the cutoff (x_res)

    :ToDo: - dynamical box construction (r_cut determines which values n1, n2, n3 have)
           - different cluster forms (spherical, cylinder, ...), add default parameters, better solution for 'orient'
    """

    # make sure we take the correct cell length for 2D structures
    if not structure.pbc[2]:
        allpos = np.array([isite.position for isite in structure.sites])
        maxpos = allpos[allpos[:, 2] == max(allpos[:, 2])]
        c3 = maxpos[0]
        c3[2] += 10  # make sure there is no overlap
        sl3 = np.sum(np.sqrt(c3**2))
    else:
        sl3 = structure.cell_lengths[2]
        c3 = structure.cell[2]

    #obtain cutoff distance from radius and h
    dist_cut = max(radius, h)

    #initialize arrays and reference the system
    x = select_reference(structure_array, i)
    x_temp = np.array([x[i]])

    #calculate needed amount of boxes in all three directions
    #========================================================
    #spherical approach (same distance in all three directions)
    if clust_shape == 'spherical':
        box_1 = int(radius / structure.cell_lengths[0] + 3)
        box_2 = int(radius / structure.cell_lengths[1] + 3)
        box_3 = int(radius / sl3 + 3)
    #cylindrical shape (different distances for the different directions)
    elif clust_shape == 'cylindrical':
        maxval = max(h / 2., radius)
        box_1 = int(maxval / structure.cell_lengths[0] + 3)
        box_2 = int(maxval / structure.cell_lengths[1] + 3)
        box_3 = int(maxval / sl3 + 3)
    else:
        raise ValueError("clust_shape needs to be either 'sperical' or 'cylindrical'")
    #================================================================================================================

    #create array of all the atoms in an expanded system
    box = max(box_1, box_2, box_3)  #pylint: disable=possibly-used-before-assignment
    cell = np.array(structure.cell)
    cell[2] = c3
    for j in range(len(x)):
        for n in range(-box, box + 1):
            for m in range(-box, box + 1):
                for l in range(-box, box + 1):
                    x_temp = np.append(
                        x_temp, [[
                            x[j][0] + (n * cell[0][0] + m * cell[1][0] + l * cell[2][0]), x[j][1] +
                            (n * cell[0][1] + m * cell[1][1] + l * cell[2][1]), x[j][2] +
                            (n * cell[0][2] + m * cell[1][2] + l * cell[2][2]), x[j][3], x[j][4], 0.
                        ]],
                        axis=0
                    )

    #x_temp now contains all the atoms and their positions regardless if they are bigger or smaller than the cutoff
    x_new = x_temp

    #calculate the distances between all the atoms and the center atom i
    for j in range(len(x_temp)):
        x_new[j][5] = get_distance(x_temp, 0, j)

    #initialize result array
    x_res = np.array([x[i]])

    #only take atoms into account whose distance to atom i is smaller than the cutoff radius
    #dist_cut = dist_cut
    if clust_shape == 'spherical':
        for j in range(len(x_temp)):
            if x_new[j][5] <= dist_cut and x_new[j][5] > 0.:
                x_res = np.append(
                    x_res, [[x_temp[j][0], x_temp[j][1], x_temp[j][2], x_temp[j][3], x_temp[j][4], x_new[j][5]]],
                    axis=0
                )
    elif clust_shape == 'cylindrical':
        for j in range(len(x_temp)):
            #rotate system into help system that is aligned with the z-axis
            x_help = rotate_onto_z(structure, x_temp, vector)

            #calculate in plane distance and vertical distance
            vert_dist = np.absolute(x_help[j][2])
            inplane_dist = np.sqrt(x_help[j][0]**2 + x_help[j][1]**2)
            #print(vert_dist, inplane_dist)
            if vert_dist <= h / 2. and inplane_dist <= radius and x_new[j][5] > 0.:
                x_res = np.append(
                    x_res, [[x_temp[j][0], x_temp[j][1], x_temp[j][2], x_temp[j][3], x_temp[j][4], x_new[j][5]]],
                    axis=0
                )

    #return an unordered array of all the atoms which are within the cutoff distance with respect to atom i
    return x_res


def write_scoef(x_res, path):
    """
    Sorts the data from find_neighbors with respect to the distance to the selected atom and writes the data
    correctly formatted into the file 'scoef'. Additionally the total number of atoms in the list is written out
    in the first line of the file.

    :param x_res: array of atoms within the cutoff radius obtained by find_neighbors (unsorted)

    :output: returns scoef file with the total number of atoms in the first line, then with the formatted positions,
             indices, charges and distances in the subsequent lines.
    """

    #write data of x_res into the 'scoef'-file
    with open_general(path, 'w') as file:
        file.write(str(f'{len(x_res):4d}'))
        file.write('\n')
        for i in range(len(x_res)):
            file.write(str(f'{x_res[i][0]:26.19e}'))
            file.write(' ')
            file.write(str(f'{x_res[i][1]:26.19e}'))
            file.write(' ')
            file.write(str(f'{x_res[i][2]:26.19e}'))
            file.write(' ')
            file.write(str(f'{int(x_res[i][3]):4d}'))
            file.write(' ')
            file.write(str(f'{x_res[i][4]:4.1f}'))
            file.write(' ')
            file.write(str(f'{x_res[i][5]:26.19e}'))
            file.write('\n')


def create_scoef_array(structure, radius, h=-1, vector=[0., 0., 1.], i=0, alat_input=None):
    """
    Creates the arrays that should be written into the 'scoef' file for a certain structure.
    Needed to conduct an impurity KKR calculation.

    :param structure: input structure of the StructureData type.
    :param radius: input cutoff radius in Ang. units.
    :param h: height of the cutoff cylinder (negative for spherical cluster shape). For negative values, clust_shape
              will be automatically assumed as 'spherical'. If there will be given a h > 0, the clust_shape
              will be 'cylindrical'.
    :param vector: orientation vector of the cylinder (just for clust_shape='cylindrical').
    :param i: atom index around which the cluster should be centered. Default: 0 (first atom in the structure).
    :param alat_input: input lattice constant in Ang. If `None` use the lattice constant that is automatically found. Otherwise rescale everything.
    """

    #shape of the cluster is specified
    if h < 0.:
        clust_shape = 'spherical'
    else:
        clust_shape = 'cylindrical'

    #store data from StructureData type in an numpy array
    structure_array = get_structure_data(structure)

    #creates an array with all the atoms within a certain cluster shape and size with respect to atom i
    c = find_neighbors(structure, structure_array, i, radius, clust_shape, h, vector)

    # convert to internal units (units of the lattice constant)
    alat = get_alat_from_bravais(np.array(structure.cell), structure.pbc[2])
    if alat_input is not None:
        # use input lattice constant instead of automatically found alat
        alat = alat_input
    # now take out alat factor
    c[:, :3] = c[:, :3] / alat  # rescale atom positions
    c[:, -1] = c[:, -1] / alat  # rescale distances

    #sort the data from c with respect to distance to the centered atom
    m = c[:, -1].argsort()
    c = c[m]

    return c


def make_scoef(structure, radius, path, h=-1., vector=[0., 0., 1.], i=0, alat_input=None):
    """
    Creates the 'scoef' file for a certain structure. Needed to conduct an impurity KKR calculation.

    :param structure: input structure of the StructureData type.
    :param radius: input cutoff radius in Ang. units.
    :param h: height of the cutoff cylinder (negative for spherical cluster shape). For negative values, clust_shape
              will be automatically assumed as 'spherical'. If there will be given a h > 0, the clust_shape
              will be 'cylindrical'.
    :param vector: orientation vector of the cylinder (just for clust_shape='cylindrical').
    :param i: atom index around which the cluster should be centered. Default: 0 (first atom in the structure).
    :param alat_input: input lattice constant in Ang. If `None` use the lattice constant that is automatically found. Otherwise rescale everything.
    """

    # returns sorted cluster array
    c = create_scoef_array(structure, radius, h, vector, i, alat_input)

    #writes out the 'scoef'-file
    write_scoef(c, path)

    return c


def write_scoef_full_imp_cls(imp_info_node, path, rescale_alat=None):
    """
    write scoef file from imp_cls info in imp_info_node
    """
    # get dictionay from node
    imp_info = imp_info_node.get_dict()

    # consistency check
    if 'imp_cls' not in list(imp_info.keys()):
        raise InputValidationError("imp_info node does not contain 'imp_cls'")

    imp_cls = np.array(imp_info.get('imp_cls'))

    # rescale if alat_input is used
    if rescale_alat is not None:
        imp_cls[:, :3] *= rescale_alat
        imp_cls[:, -1] *= rescale_alat

    # write scoef file
    write_scoef(imp_cls, path)


def get_imp_info_from_parent(parent_calc):
    """
    Returns impurity_info node from inputs to parent_calc calculation node

    Returns None if no input node of this name is found.
    """
    imp_info = None
    if 'impurity_info' in parent_calc.get_incoming().all_link_labels():
        imp_info = parent_calc.get_incoming().get_node_by_label('impurity_info')
    return imp_info



from aiida.orm import CalcJobNode, Dict, StructureData, SinglefileData, load_node, Data
from aiida.common.exceptions import InputValidationError
from aiida.common import NotExistent
from aiida_kkr.tools.find_parent import find_parent_structure, get_parent, get_remote

import os

__copyright__ = (u'Copyright (c), 2022, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.0.1'
__contributors__ = ('Markus Struckmann')


class StrucWithPotData(Data):
    def __init__(self, passedStructure=None, list_of_shapes=None, list_of_pots=None, specified_lattice_constant=False,KKRnanoCalcNode=None,\
                 VoronoiCalcNode=None, KKRhostCalcNode=None, **kwargs):
        """
        Can be created from an AiiDA structure object and a list of shapes and potentials, corresponding to this object. Also possible to create from KKRnanoCalcNode or a VoronoiCalcNode. Instead of a Voronoi calculation also a KKRhost calculation can be read in using VoronoiCalcNode
        """

        super().__init__(**kwargs)

        if not KKRnanoCalcNode == None:
            passedStructure, list_of_shapes, list_of_pots = self.get_strucwithpot_from_KKRnanoConvert(KKRnanoCalcNode)
        if not VoronoiCalcNode == None:
            passedStructure, list_of_shapes, list_of_pots = self.get_strucwithpot_from_Voronoi(VoronoiCalcNode)
        if not KKRhostCalcNode == None:
            passedStructure, list_of_shapes, list_of_pots = self.get_strucwithpot_from_Voronoi(KKRhostCalcNode)

        if not passedStructure == None and not list_of_shapes == None and not list_of_pots == None:
            self.structure = passedStructure
            self.shapes = list_of_shapes
            self.potentials = list_of_pots
        else:
            raise InputValidationError(
                'Please check input. Either a KKRnanoCalcNode or a StructureData object,\
and lists of SingleFileData potential and shape files have to be provided.'
            )

        #Possibly check if input nodes were store to the database, using e.g. node.is_stored
        if not specified_lattice_constant == False:
            self.specified_lattice_constant = specified_lattice_constant

    # Getter and Setter of the properties

    @property
    def structure(self):
        """
        The AiiDA original StructureData object
        """
        return load_node(self.get_attribute('structure'))

    @structure.setter
    def structure(self, struc):
        """
        Set the AiiDA original StructureData object.
        (Note that only the corresponding UUID is stored in the DataBase)

        :raise ValueError:
        """
        struc.store()
        self.set_attribute('structure', struc.uuid)

    @property
    def shapes(self):
        """
        The shape list in exactly the same order as stored in the StructureData object
        """
        return [load_node(item) for item in self.get_attribute('shapes')]

    @shapes.setter
    def shapes(self, shape_list):
        """
        Set the the shape list
        (Note that only the corresponding UUIDs in a list are stored in the DataBase)
        :raise ValueError:
        """
        for item in shape_list:
            item.store()
        shape_list_uuids = [item.uuid for item in shape_list]
        self.set_attribute('shapes', shape_list_uuids)

    @property
    def potentials(self):
        """
        The potential list in exactly the same order as stored in the StructureData object
        """
        return [load_node(item) for item in self.get_attribute('potentials')]

    @potentials.setter
    def potentials(self, potential_list):
        """
        Set the the potential list
        (Note that only the corresponding UUIDs in a list are stored in the DataBase)
        :raise ValueError:
        """
        for item in potential_list:
            item.store()
        potential_list_uuids = [item.uuid for item in potential_list]
        self.set_attribute('potentials', potential_list_uuids)

    @property
    def specified_lattice_constant(self):
        """
        The specified_lattice_constant to use in the input file of KKR calculations in angs
        """
        return load_node(self.get_attribute('specified_lattice_constant'))

    @specified_lattice_constant.setter
    def specified_lattice_constant(self, specified_lattice_constant):
        """
        Set the the specified_lattice_constant in angs

        :raise ValueError:
        """
        specified_lattice_constant.store()
        self.set_attribute('specified_lattice_constant', specified_lattice_constant.uuid)

    def make_shapefun(self):
        """
        Generate a shapefun file for the structure in string
        """
        #number of shapefuns in file
        shapefun = f'    {len(self.shapes)}\n'

        #header (this part is ignored)
        if len(self.shapes) % 4 == 0:
            headerlength = int(len(self.shapes) / 4)
        else:
            headerlength = int(len(self.shapes) / 4) + 1

        for k in range(headerlength):
            shapefun += '  0.100000000000E+01  0.100000000000E+01  0.100000000000E+01  0.100000000000E+01\n'

        #actual shapefuns
        for shape in self.shapes:
            shapefun += shape.get_content()

        return shapefun

    def make_potential(self):
        """
        Generate a potential file for the structure in string
        """
        potential = ''
        for vpot in self.potentials:
            potential += vpot.get_content()
        return potential

    def _check_for_CPA(self, structure, potentiallist):
        """
        returns a list of sites, which have more than one element and a potential list with tuples for theses sites
        """
        CPA_sites = []
        potential_no = 0
        new_potlist = []
        for site in structure.get_pymatgen().sites:
            no_species = len(site.species.elements)
            if no_species > 1:
                CPA_sites.append(site)
                new_potlist.append(tuple(potentiallist[potential_no:potential_no + no_species]))
                potential_no += no_species
            else:
                new_potlist.append(potentiallist[potential_no])
                potential_no += 1
        return CPA_sites, new_potlist

    def get_strucwithpot_from_Voronoi(self, calcnode):
        """
        Generate a strucwithpot object from a KKRnano calculation
        """
        #TODO: Check if input is valid

        structure = find_parent_structure(calcnode)[0]
        cwd = os.getcwd()

        #shapefun
        shapelist = []
        try:
            with calcnode.outputs.retrieved.open('shapefun') as _f:
                shapes = _f.readlines()
        except:
            calcnode.outputs.remote_folder.getfile('shapefun', f'{cwd}/shapefun')
            with open('shapefun', 'r') as reader:
                shapes = reader.readlines()
        lines = []
        for line in range(len(shapes)):
            if shapes[line].find('Shape') > 0:
                lines.append(line)
        lines.append(len(shapes))
        print('Printing out paths of shape files \n__________')
        for j in range(len(lines) - 1):
            shape_string = ''
            for k in range(lines[j], lines[j + 1]):
                shape_string += shapes[k]

            shape_no_filename = 'shape.' + str(j + 1).zfill(7)
            with open(shape_no_filename, 'w') as file:
                file.write(shape_string)
                path = os.path.realpath(file.name)

            print(path)
            abs_path = f'{cwd}/{shape_no_filename}'
            self.put_object_from_file(abs_path, shape_no_filename)  #Problem has to be called via instance
            self.set_attribute(shape_no_filename.replace('.', ''), shape_no_filename)
            with self.open(shape_no_filename, 'r') as _f:
                shapelist.append(SinglefileData(_f.name))
        print('__________')
        #potential
        potentiallist = []
        try:
            with calcnode.outputs.retrieved.open('output.pot') as _f:
                potentials = _f.readlines()
        except:
            filename = 'out_potential'
            calcnode.outputs.remote_folder.getfile(filename, f'{cwd}/{filename}')
            with open(filename, 'r') as reader:
                potentials = reader.readlines()
        lines = []
        #finding lines where the potentials start/stop
        for line in range(len(potentials)):
            if potentials[line].find('exc') > 0:
                if potentials[line].find('SPIN') > 0:
                    if potentials[line].find('SPIN DOWN') > 0:
                        lines.append(line)
                        print('Added spin-resolved potential')
                else:
                    lines.append(line)
        lines.append(len(potentials))
        print('__________')
        #creating new files with the individual potentials
        print('Printing out paths of potential files\n__________')
        for j in range(len(lines) - 1):
            potential_string = ''
            for k in range(lines[j], lines[j + 1]):
                potential_string += potentials[k]
                #potential_string+="\n"
            potential_no_filename = 'vpot.' + str(j + 1).zfill(7)
            with open(potential_no_filename, 'w') as file:
                file.write(potential_string)
                path = os.path.realpath(file.name)
            print('local path: ', path)
            abs_path = f'{cwd}/{potential_no_filename}'
            self.put_object_from_file(abs_path, potential_no_filename)  #Problem has to be called via instance
            self.set_attribute(potential_no_filename.replace('.', ''), potential_no_filename)
            with self.open(potential_no_filename, 'r') as _f:
                potentiallist.append(SinglefileData(_f.name))
                print('repository path: ', _f)
        print('__________')

        #Check for CPA
        if not len(structure.sites) == len(potentiallist):
            CPA_sites, _ = self._check_for_CPA(structure, potentiallist)

            if len(CPA_sites) == 0:
                raise InputValidationError(
                    'The number of sites in the found parent structure does not match the number of obtained potentials.'
                )
            else:
                print(
                    'WARNING: The number of sites in the found parent structure does not match the number of obtained potentials. This might be due to some CPA input which cannot be processed by KKRnano. The following sites are not (single) chemical elements: \n',
                    CPA_sites
                )

        return structure, shapelist, potentiallist

    def get_strucwithpot_from_KKRnanoConvert(self, calcnode):
        """
        Generate a strucwithpot object from a KKRnano calculation
        """
        #Check if input is valid
        if calcnode.inputs.convert.value == False:
            raise InputValidationError(
                'Only the convert step output can be processed! If this has not been done, yet, the parameter `builder.convert=Bool(False)` can be used and the process be run with 1 MPI, to obtain ASCII-potential files'
            )
        structure = find_parent_structure(calcnode)[0]
        pot_list = []
        for item in calcnode.outputs.retrieved.list_object_names():
            if item.find('vpot') == 0:
                with calcnode.outputs.retrieved.open(item) as _f:
                    path = _f.name
                pot_list.append(SinglefileData(path))

        if not len(structure.sites) == len(pot_list):
            raise InputValidationError(
                'The number of sites in the found parent structure does not match the number of obtained potentials. Found {} sites in structure, but {} potential files.'
                .format(len(structure.sites), len(pot_list))
            )

        #using the shapefun of either the input voronoi calculation or input strucwithpot object
        shape_list = []
        # best approach probably not to piggyback on the found structure and find the corresponding structure based on the first voronoi calculation that appears, or strucwithpot input
        #parent_calc=self.inputs.parent_folder#.get_incoming(node_class=CalcJobNode).first().node

        #find shapefunction
        shape_list = self.find_parent_shapefun(calcnode)

        return structure, shape_list, pot_list

    def find_parent_shapefun(self, parent_folder):
        """
        Find the shape files recursively in chain of parent calculations, either to be extracted from "shapefun" file or "shapes" files
        """
        iiter = 0
        Nmaxiter = 1000

        parent_folder_tmp = get_parent(parent_folder)

        print(parent_folder_tmp)
        parent_folder_tmp_listdir = parent_folder_tmp.listdir(
        )  #requires remote ssh connection, therefore much quicker this way
        while not 'shape.0000001' in parent_folder_tmp_listdir and not 'shapefun' in parent_folder_tmp_listdir and iiter < Nmaxiter:
            parent_folder_tmp = get_remote(get_parent(parent_folder_tmp))  #at this point the result is a CalcNode!
            parent_folder_tmp_listdir = parent_folder_tmp.outputs.remote_folder.listdir()
            iiter += 1
            print(parent_folder_tmp)
            if iiter % 200 == 0:
                print(
                    'Warning: finding shapes takes quite long (already searched {} ancestors). Stop after {}'.format(
                        iiter, Nmaxiter
                    )
                )
        if 'shape.0000001' in parent_folder_tmp_listdir or 'shapefun' in parent_folder_tmp_listdir:
            try:
                shapelist = self.extract_shapefuns(parent_folder_tmp.outputs.remote_folder)
            except AttributeError:
                shapelist = self.extract_shapefuns(parent_folder_tmp)
            return shapelist
        else:
            raise ValueError('shape files not found ')

    def extract_shapefuns(self, parent_folder):
        """
        Process shapefun files from remote folder data in the way needed for StrucWithPot
        """
        shapelist = []
        cwd = os.getcwd()
        print('Extracting shapes')
        has_shapefun_file = False
        found_shapes = False  #to see if KKRnano produced some shapes as output (better than using shapefun)
        parent_folder_listdir = parent_folder.listdir()  #quicker this way as based on SSH tunneling
        for filename in parent_folder_listdir:
            if filename.find('shape.') >= 0:
                if has_shapefun_file:
                    shapelist = []
                abs_path = f'{cwd}/{filename}'
                parent_folder.getfile(filename, f'{cwd}/{filename}')
                self.put_object_from_file(abs_path, filename)
                self.set_attribute(filename.replace('.', ''), filename)
                os.remove(filename)
                with self.open(filename, 'r') as _f:
                    shapelist.append(SinglefileData(_f.name))
                    print('Found shape in repsitory:')
                    print(_f.name)
                has_shapefun_file = False
                found_shapes = True
        if 'shapefun' in parent_folder_listdir and not found_shapes:
            filename = 'shapefun'
            print('Shapefun in dir, this part of the program might need more testing')
            abs_path = f'{cwd}/{filename}'
            parent_folder.getfile(filename, f'{cwd}/{filename}')

            with open(filename, 'r') as reader:
                shapes = reader.readlines()


#                 print(os.path.realpath(reader.name))
            lines = []
            for line in range(len(shapes)):
                if shapes[line].find('Shape') > 0:
                    lines.append(line)
            lines.append(len(shapes))
            for j in range(len(lines) - 1):
                shape_string = ''
                for k in range(lines[j], lines[j + 1]):
                    shape_string += shapes[k]

                shape_no_filename = 'shape.' + str(j + 1).zfill(7)
                with open(shape_no_filename, 'w') as file:
                    file.write(shape_string)
                    path = os.path.realpath(file.name)
                print(path)
                abs_path = f'{cwd}/{shape_no_filename}'
                self.put_object_from_file(abs_path, shape_no_filename)  #Problem has to be called via instance
                self.set_attribute(shape_no_filename.replace('.', ''), shape_no_filename)
                with self.open(shape_no_filename, 'r') as _f:
                    shapelist.append(SinglefileData(_f.name))
                os.remove(shape_no_filename)
            has_shapefun_file = True
        if has_shapefun_file:
            print(
                'WARNING: Only a shapefun from some Voronoi input was found, it is possible that the potential does not match the shapefun parameters, unless they are set this way explicitly in the respective input file! It is advisable to use the `write_shapes=1` command in input.conf'
            )
        print('Found shapelist:')
        print(shapelist)
        return shapelist

    def sites(self):
        """
        Gives back a list of sites. Particularly useful for processing CPA input.
        (AiiDA structure site, potential file(s), shapefun)
        """
        struc = self.structure
        pots = self.potentials
        shapes = self.shapes

        _, new_pots = self._check_for_CPA(struc, pots)
        site_list = []
        for j in range(len(struc.sites)):
            site_list.append({
                'StructureDataSite': struc.sites[j],
                'potential': new_pots[j],
                'shapefun': shapes[j % len(shapes)]
            })
        return site_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In this module you find the base workflow for a dos calculation and
some helper methods to do so with AiiDA
"""

from aiida import orm
from masci_tools.io.kkr_params import kkrparams
from aiida.engine import WorkChain, if_, ToContext
from aiida.engine import submit
from aiida_kkr.tools.common_workfunctions import (
    test_and_get_codenode,
    get_parent_paranode,
    update_params_wf,
    get_inputs_kkr,
)
from aiida_kkr.calculations.kkr import KkrCalculation
from aiida_kkr.calculations.voro import VoronoiCalculation
from aiida.common.exceptions import InputValidationError
from aiida_kkr.tools.parse_dos import parse_dosfiles
from aiida_kkr.tools.save_output_nodes import create_out_dict_node
from aiida_kkr.tools.extract_kkrhost_noco_angles import extract_noco_angles
from aiida_kkr.workflows.bs import set_energy_params

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.8.4'
__contributors__ = u'Philipp Rüßmann'


class kkr_dos_wc(WorkChain):
    """
    Workchain a DOS calculation with KKR starting from the remoteData node
    of a previous calculation (either Voronoi or KKR).

    :param wf_parameters: (Dict); Workchain specifications
    :param options: (Dict); specifications for the computer
    :param remote_data: (RemoteData), mandatory; from a KKR or Vornoi calculation
    :param kkr: (Code), mandatory; KKR code running the dos calculation

    :return result_kkr_dos_wc: (Dict), Information of workflow results
        like Success, last result node, list with convergence behavior
    """

    _workflowversion = __version__
    _wf_label = 'kkr_dos_wc'
    _wf_description = 'Workflow for a KKR dos calculation starting either from a structure with automatic voronoi calculation or a valid RemoteData node of a previous calculation.'
    _wf_default = {
        'nepts': 96,  # number of points in contour
        'tempr': 200.,  # K, smearing temperature
        'emin': -10.,  # eV, rel to EF, start of energy contour
        'emax': 5.,  # eV       end of energy contour
        'kmesh': [30, 30, 30],  # kmesh for DOS calculation (typically higher than in scf contour)
        'RCLUSTZ': None,  # cluster radiu, only used if a value is set
    }
    _options_default = {
        'queue_name': '',  # Queue name to submit jobs to
        # resources to allowcate for the job
        'resources': {
            'num_machines': 1
        },
        # walltime after which the job gets killed (gets parsed to KKR)
        'max_wallclock_seconds': 60 * 60,
        'withmpi': True,  # execute KKR with mpi or without
        'custom_scheduler_commands': '',  # some additional scheduler commands
        'prepend_text': '',
        'append_text': '',
        'additional_retrieve_list': None
    }

    # intended to guide user interactively in setting up a valid wf_params node
    @classmethod
    def get_wf_defaults(cls, silent=False):
        """
        Print and return _wf_defaults dictionary.

        Can be used to easily create set of wf_parameters.
        returns _wf_defaults
        """
        if not silent:
            print(f'Version of workflow: {cls._workflowversion}')
        return cls._wf_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Defines the outline of the workflow.
        """
        # Take input of the workflow or use defaults defined above
        super(kkr_dos_wc, cls).define(spec)
        spec.input(
            'wf_parameters',
            valid_type=orm.Dict,
            required=False,
            default=lambda: orm.Dict(dict=cls._wf_default),
            help='Workflow parameter (see `kkr_dos_wc.get_wf_defaults()`).'
        )
        spec.input(
            'options',
            valid_type=orm.Dict,
            required=False,
            default=lambda: orm.Dict(dict=cls._wf_default),
            help='Computer options used by the workflow.'
        )
        spec.input(
            'remote_data', valid_type=orm.RemoteData, required=True, help='RemoteData node of the parent calculation.'
        )
        spec.input('kkr', valid_type=orm.Code, required=True, help='KKRhost Code node used to run the DOS calculation.')
        spec.input(
            'initial_noco_angles',
            valid_type=orm.Dict,
            required=False,
            help="""Initial non-collinear angles for the magnetic moments. See KkrCalculation for details.
            If this is found in the input potentially extracted nonco angles from the parent calulation are overwritten!"""
        )
        # maybe overwrite some settings from the KKRhost convergence run
        spec.input(
            'params_kkr_overwrite',
            valid_type=orm.Dict,
            required=False,
            help='Overwrite some input parameters of the parent KKR calculation.'
        )
        # expose LDAU input node
        spec.input(
            'settings_LDAU',
            valid_type=orm.Dict,
            required=False,
            help="""
Settings for running a LDA+U calculation. The Dict node should be of the form
    settings_LDAU = Dict(dict={'iatom=0':{
        'L': 3,         # l-block which gets U correction (1: p, 2: d, 3: f-electrons)
        'U': 7.,        # U value in eV
        'J': 0.75,      # J value in eV
        'Eref_EF': 0.,  # reference energy in eV relative to the Fermi energy. This is the energy where the projector wavefunctions are calculated (should be close in energy where the states that are shifted lie (e.g. for Eu use the Fermi energy))
    }})
    Note: you can add multiple entries like the one for iatom==0 in this example. The atom index refers to the corresponding atom in the impurity cluster.
"""
        )

        # define outputs
        spec.output('results_wf', valid_type=orm.Dict, required=True, help='Results collected by the workflow.')
        spec.output('dos_data', valid_type=orm.XyData, required=False, help='XyData node of the parsed DOS output.')
        spec.output(
            'dos_data_interpol',
            valid_type=orm.XyData,
            required=False,
            help='XyData node of the parsed DOS output, interpolated onto the real axis.'
        )

        # Here the structure of the workflow is defined
        spec.outline(
            # initialize workflow
            cls.start,
            # validate input
            if_(cls.validate_input)(
                # set DOS contour in parameter node
                cls.set_params_dos,
                # calculate DOS and interpolate result
                # TODO: encapsulate get_dos in restarting mechanism (should be a base class of workflows that start calculations)
                # i.e. use base_restart_calc workchain as parent
                cls.get_dos
            ),
            #  collect results and store DOS output as ArrayData node
            # (dos, lmdos, dos.interpolate, ...)
            cls.return_results
        )

        # definition of exit code in case something goes wrong in this workflow
        spec.exit_code(
            161,
            'ERROR_NO_INPUT_REMOTE_DATA',
            'No remote_data was provided as Input',
        )
        spec.exit_code(
            162,
            'ERROR_KKRCODE_NOT_CORRECT',
            'The code you provided for kkr does not use the plugin kkr.kkr',
        )
        spec.exit_code(
            163,
            'ERROR_CALC_PARAMETERS_INVALID',
            'calc_parameters given are not consistent! Hint: did you give an unknown keyword?',
        )
        spec.exit_code(
            164,
            'ERROR_CALC_PARAMETERS_INCOMPLETE',
            'calc_parameters not complete',
        )
        spec.exit_code(
            165,
            'ERROR_DOS_PARAMS_INVALID',
            'dos_params given in wf_params are not valid',
        )
        spec.exit_code(
            166,
            'ERROR_DOS_CALC_FAILED',
            'KKR dos calculation failed',
        )

    def start(self):
        """
        init context and some parameters
        """
        self.report(f'INFO: started KKR dos workflow version {self._workflowversion}')

        # input para
        wf_dict = self.inputs.wf_parameters.get_dict()
        options_dict = self.inputs.options.get_dict()

        # TODO: check for completeness
        if wf_dict == {}:
            wf_dict = self._wf_default
            self.report('INFO: using default wf parameter')
        if options_dict == {}:
            options_dict = self._options_default
            self.report('INFO: using default options')
        self.ctx.append_text = options_dict.get('append_text', self._options_default['append_text'])
        self.ctx.prepend_text = options_dict.get('prepend_text', self._options_default['prepend_text'])
        self.ctx.additional_retrieve_list = options_dict.get(
            'additional_retrieve_list', self._options_default['additional_retrieve_list']
        )

        # set values, or defaults
        self.ctx.withmpi = options_dict.get(
            'withmpi',
            self._options_default['withmpi'],
        )
        self.ctx.resources = options_dict.get(
            'resources',
            self._options_default['resources'],
        )
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds',
            self._options_default['max_wallclock_seconds'],
        )
        self.ctx.queue = options_dict.get(
            'queue_name',
            self._options_default['queue_name'],
        )
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands',
            self._options_default['custom_scheduler_commands'],
        )

        self.ctx.dos_params_dict = self.inputs.wf_parameters.get_dict()
        self.ctx.dos_kkrparams = None  # is set in set_params_dos

        self.ctx.description_wf = self.inputs.get(
            'description',
            self._wf_description,
        )
        self.ctx.label_wf = self.inputs.get(
            'label',
            self._wf_label,
        )

        self.report(
            f'INFO: use the following parameter:\n'
            f'withmpi: {self.ctx.withmpi}\n'
            f'Resources: {self.ctx.resources}\n'
            f'Walltime (s): {self.ctx.max_wallclock_seconds}\n'
            f'queue name: {self.ctx.queue}\n'
            f'scheduler command: {self.ctx.custom_scheduler_commands}\n'
            f'description: {self.ctx.description_wf}\n'
            f'label: {self.ctx.label_wf}\n'
            f'dos_params: {self.ctx.dos_params_dict}\n'
        )

        # return para/vars
        self.ctx.successful = True
        self.ctx.errors = []
        self.ctx.formula = ''

    def validate_input(self):
        """
        # validate input and find out which path (1, or 2) to take
        # return True means run voronoi if false run kkr directly
        """
        inputs = self.inputs

        if 'remote_data' in inputs:
            input_ok = True
        else:
            input_ok = False
            return self.exit_codes.ERROR_NO_INPUT_REMOTE_DATA  # pylint: disable=no-member

        # extract correct remote folder of last calculation if input
        # remote_folder node is not from KkrCalculation but kkr_scf_wc workflow
        input_remote = self.inputs.remote_data
        # check if input_remote has single KkrCalculation parent
        parents = input_remote.get_incoming(node_class=orm.CalcJobNode)
        nparents = len(parents.all_link_labels())
        if nparents != 1:
            # extract parent workflow and get uuid of last calc from output node
            parent_workflow = input_remote.inputs.last_RemoteData
            if not isinstance(parent_workflow, orm.WorkChainNode):
                raise InputValidationError(
                    'Input remote_data node neither output of a KKR/voronoi calculation nor of kkr_scf_wc workflow'
                )
            parent_workflow_out = parent_workflow.outputs.output_kkr_scf_wc_ParameterResults
            uuid_last_calc = parent_workflow_out.get_dict().get('last_calc_nodeinfo').get('uuid')
            last_calc = orm.load_node(uuid_last_calc)
            if not isinstance(last_calc, KkrCalculation) and not isinstance(last_calc, VoronoiCalculation):
                raise InputValidationError(
                    'Extracted last_calc node not of type KkrCalculation: check remote_data input node'
                )
            # overwrite remote_data node with extracted remote folder
            output_remote = last_calc.outputs.remote_folder
            self.inputs.remote_data = output_remote

        if 'kkr' in inputs:
            try:
                test_and_get_codenode(inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                input_ok = False
                return self.exit_codes.ERROR_KKRCODE_NOT_CORRECT  # pylint: disable=no-member

        # set self.ctx.input_params_KKR
        self.ctx.input_params_KKR = get_parent_paranode(self.inputs.remote_data)

        return input_ok

    def set_params_dos(self):
        """
        take input parameter node and change to DOS contour according to input from wf_parameter input
        internally calls the update_params work function to keep track of provenance
        """
        params = self.ctx.input_params_KKR
        input_dict = params.get_dict()
        para_check = kkrparams()

        # step 1 try to fill keywords
        try:
            for key, val in input_dict.items():
                para_check.set_value(key, val, silent=True)
        except:
            return self.exit_codes.ERROR_CALC_PARAMETERS_INVALID  # pylint: disable=no-member

        # step 2: check if all mandatory keys are there
        label = ''
        descr = ''
        missing_list = para_check.get_missing_keys(use_aiida=True)
        if missing_list != []:
            kkrdefaults = kkrparams.get_KKRcalc_parameter_defaults()[0]
            kkrdefaults_updated = []
            for key_default, val_default in list(kkrdefaults.items()):
                if key_default in missing_list:
                    para_check.set_value(key_default, kkrdefaults.get(key_default), silent=True)
                    kkrdefaults_updated.append(key_default)
                    missing_list.remove(key_default)
            if len(missing_list) > 0:
                self.report(f'ERROR: calc_parameters misses keys: {missing_list}')
                return self.exit_codes.ERROR_CALC_PARAMETERS_INCOMPLETE  # pylint: disable=no-member
            self.report(f'updated KKR parameter node with default values: {kkrdefaults_updated}')
            label = 'add_defaults_'
            descr = 'added missing default keys, '

        # overwrite energy contour to DOS contour no matter what is in input parameter node.
        # Contour parameter given as input to workflow.
        econt_new = self.ctx.dos_params_dict
        # always overwrite NPOL, N1, N3, thus add these to econt_new
        econt_new['NPOL'] = 0
        econt_new['NPT1'] = 0
        econt_new['NPT3'] = 0
        parent_calc = self.inputs.remote_data.get_incoming(node_class=orm.CalcJobNode).first().node
        if parent_calc.process_label == 'VoronoiCalculation':
            # for the voronoi calculation we need to calculate the Fermi level since it is not in the output parameters directly
            voro_out_para = parent_calc.outputs.output_parameters.get_dict()
            ef = voro_out_para['emin'] - voro_out_para['emin_minus_efermi_Ry']
        else:
            ef = parent_calc.outputs.output_parameters.get_dict()['fermi_energy']  # unit in Ry
        try:
            para_check = set_energy_params(econt_new, ef, para_check)
        except:
            return self.exit_codes.ERROR_DOS_PARAMS_INVALID  # pylint: disable=no-member

        updatenode = orm.Dict(dict=para_check.get_dict())
        updatenode.label = label + 'KKRparam_DOS'
        updatenode.description = descr + \
            'KKR parameter node extracted from parent parameters and wf_parameter input node.'

        paranode_dos = update_params_wf(self.ctx.input_params_KKR, updatenode)

        # maybe overwrite some inputs
        if 'params_kkr_overwrite' in self.inputs:
            self.report(f'found params_kkr_overwrite: {self.inputs.params_kkr_overwrite.get_dict()}')
            updatenode = self.inputs.params_kkr_overwrite
            updatenode.label = 'params overwrite'
            paranode_dos = update_params_wf(paranode_dos, updatenode)

        self.ctx.dos_kkrparams = paranode_dos

    def get_dos(self):
        """
        submit a dos calculation and interpolate result if returns complete
        """

        label = 'KKR DOS calc.'
        dosdict = self.ctx.dos_params_dict
        description = 'dos calc: emin= {}, emax= {}, nepts= {}, tempr={}, kmesh={}'.format(
            dosdict['emin'], dosdict['emax'], dosdict['nepts'], dosdict['tempr'], dosdict['kmesh']
        )
        code = self.inputs.kkr
        remote = self.inputs.remote_data
        params = self.ctx.dos_kkrparams
        options = {
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'resources': self.ctx.resources,
            'queue_name': self.ctx.queue
        }
        if self.ctx.custom_scheduler_commands:
            options['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        if self.ctx.append_text:
            options['append_text'] = self.ctx.append_text
        if self.ctx.prepend_text:
            options['prepend_text'] = self.ctx.prepend_text
        if self.ctx.additional_retrieve_list:
            options['additional_retrieve_list'] = self.ctx.additional_retrieve_list

        inputs = get_inputs_kkr(
            code, remote, options, label, description, parameters=params, serial=(not self.ctx.withmpi)
        )

        # add nonco angles if found in the parent calculation or in the input
        if 'initial_noco_angles' in self.inputs:
            # overwrite nonco_angles from the input if given
            inputs['initial_noco_angles'] = self.inputs.initial_noco_angles
            self.report('used nonco angles from input to workflow')
        else:
            # extract from the parent calculation
            parent_calc = remote.get_incoming(node_class=orm.CalcJobNode).first().node
            if 'initial_noco_angles' in parent_calc.inputs:
                noco_angles = extract_noco_angles(
                    fix_dir_threshold=orm.Float(1e-6),  # make small enough
                    old_noco_angles=parent_calc.inputs.initial_noco_angles,
                    last_retrieved=parent_calc.outputs.retrieved
                )
                # set nonco angles (either from input or from output if it was updated)
                self.report(noco_angles)
                if noco_angles == {}:
                    noco_angles = parent_calc.inputs.initial_noco_angles
                inputs['initial_noco_angles'] = noco_angles
                self.report(f'extract nonco angles and use from parent ({noco_angles})')

        # LDA+U settings
        if 'settings_LDAU' in self.inputs:
            self.report('Add settings_LDAU input node')
            inputs.settings_LDAU = self.inputs.settings_LDAU

        # run the DOS calculation
        self.report('INFO: doing calculation')
        dosrun = self.submit(KkrCalculation, **inputs)

        # for restart workflow:
        self.ctx.last_calc = dosrun

        return ToContext(dosrun=dosrun)

    def return_results(self):
        """
        Collect results, parse DOS output and link output nodes to workflow node
        """

        # check wether or not calculation was taken from cached node
        caching_info = f'INFO: cache_source of dos calc node: {self.ctx.dosrun.get_cache_source()}'
        print(caching_info)
        self.report(caching_info)

        # capture error of unsuccessful DOS run
        if not self.ctx.dosrun.is_finished_ok:
            self.ctx.successful = False
            error = (f'ERROR: DOS calculation failed somehow it is in state {self.ctx.dosrun.process_state}')
            print(error)
            from pprint import pprint
            pprint(self.ctx.dosrun.attributes)
            print('stdout', self.ctx.dosrun.get_scheduler_stdout())
            print('stderr', self.ctx.dosrun.get_scheduler_stderr())
            self.report(error)
            self.ctx.errors.append(error)
            return self.exit_codes.ERROR_DOS_CALC_FAILED  # pylint: disable=no-member

        # create dict to store results of workflow output
        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['withmpi'] = self.ctx.withmpi
        outputnode_dict['resources'] = self.ctx.resources
        outputnode_dict['max_wallclock_seconds'] = self.ctx.max_wallclock_seconds
        outputnode_dict['queue_name'] = self.ctx.queue
        outputnode_dict['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        outputnode_dict['dos_params'] = self.ctx.dos_params_dict
        try:
            outputnode_dict['nspin'] = self.ctx.dosrun.res.nspin
        except:
            error = 'ERROR: nspin not extracted'
            self.report(error)
            self.ctx.successful = False
            self.ctx.errors.append(error)
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['list_of_errors'] = self.ctx.errors

        # create output node with data-provenance
        outputnode = orm.Dict(dict=outputnode_dict)
        outputnode.label = 'kkr_scf_wc_results'
        outputnode.description = ''

        self.report('INFO: create dos results nodes')
        try:
            self.report(f'INFO: create dos results nodes. dos calc retrieved node={self.ctx.dosrun.outputs.retrieved}')
            has_dosrun = True
        except AttributeError as _error:
            self.report('ERROR: no dos calc retrieved node found')
            self.report(f'Caught AttributeError {_error}')
            has_dosrun = False

        # interpol dos file and store to XyData nodes
        if has_dosrun:
            dos_retrieved = self.ctx.dosrun.outputs.retrieved
            if 'complex.dos' in dos_retrieved.list_object_names():
                dosXyDatas = parse_dosfiles(dos_retrieved)

        # collect output nodes with link labels
        outdict = {}
        if has_dosrun:
            outdict['dos_data'] = dosXyDatas['dos_data']
            outdict['dos_data_interpol'] = dosXyDatas['dos_data_interpol']
        # create data provenance of results node
        link_nodes = outdict.copy()  # link also dos output nodes
        if has_dosrun:
            link_nodes['doscalc_remote'] = self.ctx.dosrun.outputs.remote_folder
        outdict['results_wf'] = create_out_dict_node(outputnode, **link_nodes)

        # create links to output nodes
        for link_name, node in outdict.items():
            self.out(link_name, node)

        self.report('INFO: done with DOS workflow!\n')



# -*- coding: utf-8 -*-
"""
This module contains helper functions and tools doing STM-like scans around impurity clusters
"""

import numpy as np
from aiida import orm, engine
from aiida_kkr.tools import find_parent_structure
from aiida_kkr.tools.combine_imps import get_scoef_single_imp
from aiida_kkr.tools.imp_cluster_tools import pos_exists_already, combine_clusters
from masci_tools.io.common_functions import get_alat_from_bravais

__copyright__ = (u'Copyright (c), 2023, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.4'
__contributors__ = (u'Philipp Rüßmann', u'Raffaele Aliberti')

##############################################################################
# combine impurty clusters


def convert_to_imp_cls(host_structure, imp_info):
    """
    convert imp info to rcls form
    """
    if 'imp_cls' in imp_info.get_dict():
        clust1 = np.array(imp_info['imp_cls'])
        imp_info_cls = imp_info
    else:
        # convert Zimp, Rcut info to imp_cls info
        clust1 = get_scoef_single_imp(host_structure, imp_info)
        imp_info_cls = orm.Dict({'imp_cls': clust1, 'Zimp': imp_info['Zimp'], 'Rimp_rel': [[0., 0., 0.]]})
    return imp_info_cls, clust1


def get_Zadd(host_structure, add_position):
    """
    get Zatom for adding position
    """
    from aiida.common.constants import elements as PeriodicTableElements

    _atomic_numbers = {data['symbol']: num for num, data in PeriodicTableElements.items()}

    kind_name = host_structure.sites[add_position['ilayer']].kind_name
    symbol = host_structure.get_kind(kind_name).symbol
    Zadd = float(_atomic_numbers[symbol])
    return Zadd


def get_imp_cls_add(host_structure, add_position):
    """
    define auxiliary imp_info for adding position and generate rcls
    """
    Zadd = get_Zadd(host_structure, add_position)
    ilayer = add_position['ilayer']
    imp_info2 = orm.Dict({'ilayer_center': ilayer, 'Zimp': [Zadd], 'Rcut': 1e-5})
    # old version is too slow:
    #clust2 = get_scoef_single_imp(host_structure, imp_info2)
    # new version creates the array without calling the get_scoef_single_imp function:
    clust2 = np.array([[0., 0., 0., ilayer + 1, 0., 0.]])
    return imp_info2, clust2


def get_r_offset(clust1, clust2, host_structure, add_position):
    """
    find offset vector in rcls units
    """
    # calculate out-of plane vector from the ilayer indices of the two clusters
    r_out_of_plane = np.array([0., 0., 0.])
    # minus 1 because of conversion from fortran to python standard (counting starts at 0)
    ilayer1 = int(clust1[0, 3]) - 1
    ilayer2 = int(clust2[0, 3]) - 1
    if ilayer1 != ilayer2:
        pos1 = np.array(host_structure.sites[ilayer1].position)
        pos2 = np.array(host_structure.sites[ilayer2].position)
        r_out_of_plane = pos2 - pos1

    # convert from Ang to alat units (used internally in KKR)
    alat = get_alat_from_bravais(np.array(host_structure.cell), host_structure.pbc[2])
    r_out_of_plane /= alat

    # # remove spurious offsets that might be there due to the choice of the unit cell positions
    # r_out_of_plane = np.round(r_out_of_plane, 7)
    # r_out_of_plane[:2] %= 1  # modulo 1 for x and y coordinate

    # calculate in-plane vector from da, db inputs
    da = add_position.get_dict().get('da', 0)
    db = add_position.get_dict().get('db', 0)
    cell = np.array(host_structure.cell)
    r_in_plane = da * cell[0] + db * cell[1]

    # convert from Ang to alat units (used internally in KKR)
    r_in_plane /= alat

    # combine to offset vector
    r_offset = r_out_of_plane + r_in_plane

    return r_offset


def offset_clust2(clust1, clust2, host_structure, add_position):
    """
    Compute and add offset to clust2
    """
    r_offset = get_r_offset(clust1, clust2, host_structure, add_position)

    clust2_offset = clust2.copy()
    clust2_offset[:, :3] += r_offset

    return clust2_offset


def get_imp_info_add_position(add_position, host_structure, imp_info):
    """
    Create combined impurity info node for the original
    imp cluster + an additional (STM tip) position
    """

    # extract host structure
    # host_structure = find_parent_structure(host_calc)

    # convert imp info to cls form
    imp_info_cls, clust1 = convert_to_imp_cls(host_structure, imp_info)

    # get imp cluster for adding position
    imp_info2, clust2 = get_imp_cls_add(host_structure, add_position)

    # shift clust2 by offset
    clust2_offset = offset_clust2(clust1, clust2, host_structure, add_position)

    # combine cluster information
    pos_exists_in_imp1, _ = pos_exists_already(clust1, clust2)
    if pos_exists_in_imp1:
        # If the position exists already we simply skip the addition of the new scanning position
        return None
        #raise ValueError('Additional position exists already in impurity cluster.')
    cluster_combined, rimp_rel_combined, _, _ = combine_clusters(clust1, clust2_offset, False, debug=False)
    # combine the zimp arrays
    zimp1 = imp_info['Zimp']
    if not isinstance(zimp1, list):
        # convert to list if necessary
        zimp1 = [zimp1]
    zimp_combined = zimp1 + imp_info2['Zimp']
    # now combine the imp info node
    imp_info_combined = orm.Dict({'imp_cls': cluster_combined, 'Zimp': zimp_combined, 'Rimp_rel': rimp_rel_combined})

    return imp_info_combined


@engine.calcfunction
def get_imp_info_add_position_cf(add_position, host_structure, imp_info):
    """
    Create a new impurty info node that combines the impurity cluster
    of an original calculation and an STM scanning position.
    """

    # then combine the imp info
    imp_info_combined = get_imp_info_add_position(add_position, host_structure, imp_info)

    return imp_info_combined


##############################################################################
# combine potentials


def extract_host_potential(add_position, host_calc):
    """
    Extract the potential of the position in the host that matches the additional position
    """

    # find ilayer from input node
    ilayer = add_position['ilayer']

    # get host calculation from remote
    #host_calc = host_remote.get_incoming(node_class=orm.CalcJobNode).first().node

    # read potential from host's retrieved node
    with host_calc.outputs.retrieved.open('out_potential') as _f:
        pot_txt = _f.readlines()
    iline_startpot = np.array([i for i, l in enumerate(pot_txt) if 'exc:' in l])

    # extract nspin from host calc
    nspin = host_calc.inputs.parameters['NSPIN']

    # get host's potential from ilayer
    pot_add = []
    for ispin in range(nspin):
        istart = iline_startpot[ilayer * nspin + ispin]
        iend = iline_startpot[ilayer * nspin + ispin + 1]
        pot_add += pot_txt[istart:iend]

    return pot_add


def add_host_potential_to_imp(add_position, host_calc, imp_potential_node):
    """
    combine host potential with impurity potential
    """
    # get add potential from host
    potname = f'host_pot:{add_position["ilayer"]}'
    if potname in imp_potential_node.extras:
        # reuse existing host position if we have found it previously
        pot_add = imp_potential_node.extras[potname]
    else:
        # get host postition and store as extra
        pot_add = extract_host_potential(add_position, host_calc)
        imp_potential_node.set_extra(potname, pot_add)

    # get impurity potential and convert to list
    pot_imp = imp_potential_node.get_content().split('\n')
    pot_imp = [line + '\n' for line in pot_imp if line != '']

    # glue potentials together and create SinglefileData
    pot_combined = pot_imp + pot_add

    return pot_combined


def create_combined_potential_node(add_position, host_calc, imp_potential_node):
    """
    Combine impurity potential with an additional potential from the host for
    the STM tip position (additional position)
    """
    import io

    # combine potential texts
    pot_combined = add_host_potential_to_imp(add_position, host_calc, imp_potential_node)

    # convert to byte string and put into SinglefilData node
    pot_combined_string = ''
    for line in pot_combined:
        pot_combined_string += line
    pot_combined_node = orm.SinglefileData(io.BytesIO(bytes(pot_combined_string, 'utf8')))

    return pot_combined_node


@engine.calcfunction
def create_combined_potential_node_cf(add_position, host_calc, imp_potential_node):
    """
    Calcfunction that combines the impurity potential with an addition potential site from the host
    """

    pot_combined_node = create_combined_potential_node(add_position, host_calc, imp_potential_node)

    return pot_combined_node


#####################################################################
# STM pathfinder


def STM_pathfinder(host_remote):
    """
    Calcfunction that gives back the structural information of the film, and the symmetries of the system

    inputs ::

           host_remote : Remote_data : node containing the remote data of the host material

    return ::

           plane_vectors   : list : list containing the 2 in plane vectors that span the surface.
           unique_matrices : list : list of matrices, contains the 2x2 matrices that constitue the symmetry operations of the system.
    """

    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer, SymmOp

    struc = find_parent_structure(host_remote)
    # clone the structure since it has already been saved in AiiDA and cannot be modified
    supp_struc = struc.clone()

    # If the structure is not periodic in every direction we force it to be.
    supp_struc.pbc = (True, True, True)

    # Pymatgen struc
    py_struc = supp_struc.get_pymatgen()

    struc_dict = py_struc.as_dict()
    # Find the Bravais vectors that are in-plane vectors (assumes 2D structure)
    plane_vectors = {'plane_vectors': [], 'space_group': ''}
    for vec in struc_dict['lattice']['matrix']:
        # Is this sufficient to find all the in-plane vectors?
        if vec[2] == 0 or (struc.pbc[2] and (vec[0] + vec[1]) > 0):
            plane_vectors['plane_vectors'].append(vec[:2])
    # finally check if setting of plane_vectors worked
    if 'plane_vectors' not in plane_vectors:
        raise ValueError('Could not set "plane_vectors" in STM_pathfinder')

    # Here we get the symmetry operations that are possible
    symmetry_matrices = SpacegroupAnalyzer(py_struc).get_point_group_operations(cartesian=True)

    plane_vectors['space_group'] = SpacegroupAnalyzer(py_struc).get_symmetry_dataset()['number']

    # Here we get the symmetry rotations

    supp_mat = []

    # Get the affine representation of the matrices
    for symmops in range(len(symmetry_matrices)):
        supp_mat.append(np.array(SymmOp.as_dict(symmetry_matrices[symmops])['matrix'][:3]))

    # Get only the rotation matrices and correct the numerical error
    rot_mat = []

    # Take only the matrices for the in-plane rotations
    for elements in range(len(supp_mat)):
        rot_mat.append(supp_mat[elements][0:2, 0:2])

    # Sometimes it will happen that some rotation matrices projected to the 2D space will have the same representation.
    # here we only take the unique ones
    unique_matrices = []
    for matrix in rot_mat:
        if not any(np.array_equal(matrix, m) for m in unique_matrices):
            unique_matrices.append(matrix)

    # Round off the numerical error
    for elements in range(len(unique_matrices)):
        for rows in range(len(unique_matrices[elements])):
            for cols in range(len(unique_matrices[elements][rows])):
                unique_matrices[elements][rows][cols] = round(unique_matrices[elements][rows][cols])

    return plane_vectors, unique_matrices


@engine.calcfunction
def STM_pathfinder_cf(host_structure):
    """
    Calcfunction that gives back the structural information of the film, and the symmetries of the system
    """

    struc_info, symm_matrices = STM_pathfinder(host_structure)

    return struc_info, symm_matrices


###################################################################
# lattice generation (function of lattice plot)


def lattice_generation(rot, vec, x_start, y_start, xmax, ymax):
    """

        inputs ::

        x_len  : int  : value to create points between - x and x.
        y_len  : int  : value to create points between - y and y.
        rot    : list : list of the rotation matrices given by the symmetry of the system.
        vec    : list : list containing the two Bravais vectors.
        start_x: int  : starting value for the lattice generation in the x direction
        start_y: int  : starting value for the lattice generation in the y direction

        return ::

        points_to_eliminate : list : list of list containing the (x,y) positions to NOT to
                                     be scanned
        points_to_scan      : list : list of list containing the (x,y) positions to BE be
                                     scanned
        """

    # Here we create a grid  made of points which are the linear combination of the lattice vectors
    x_len = xmax * 10
    y_len = ymax * 10  # maybe there is a way to make this more efficient...

    x_interval = [i for i in range(-x_len, x_len)]

    y_interval = [i for i in range(-y_len, y_len)]

    points_to_scan = []

    # Now we only generate points in the first quadrant and the we use the symmetry analysis
    # To visualize the other unscanned sites

    for i in x_interval:
        points_to_scan_col = []
        for j in y_interval:
            p = [i * x + j * y for x, y in zip(vec[0], vec[1])]
            if ((p[0] < 0 or p[0] > xmax) or (p[1] < 0 or p[1] > ymax)) or (p[0] < x_start or p[1] < y_start):
                continue
            else:
                points_to_scan_col.append(p)
        if len(points_to_scan_col) != 0:
            points_to_scan.append(points_to_scan_col)

    #print(lattice_points)
    points_to_eliminate = []

    for i in range(len(points_to_scan)):
        for j in range(len(points_to_scan[i])):
            #print(lattice_points[i][j])
            #if lattice_points[i][j][0] >= 0 and lattice_points[i][j][1] >= 0:
            for element in rot[1:]:
                point = np.dot(element, points_to_scan[i][j])
                #if point[0] >= 0 and point[1] >=0:
                #    continue
                #else:
                points_to_eliminate.append(point)

    #print(point_to_eliminate)

    return points_to_eliminate, points_to_scan


###############################################################################################
# lattice plot


def lattice_plot(plane_vectors, symm_vec, symm_matrices, grid_length_x, grid_length_y):
    """
        Helper tool to plot the position that will be scanned in the submission of the
        kkr_STM_wc workchain

        inputs ::

        plane_vectors : list : list containing the Bravais vector of the 2D lattice
        symm_vec      : bool : Toggle to show or not the Bravais vectors in the plot
        symm_matrices : list : list of the point-group symmetries matrices of the system
        grid_length_x : int  : scanning distance in the x direction
        grid_length_y : int  : scanning distance in the y direction

        return ::

        None

    """

    #from aiida_kkr.tools.tools_STM_scan import lattice_generation
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D

    origin = np.array([[0, 0], [0, 0]])
    # Generation of the points to plot
    unused, used = lattice_generation(symm_matrices, plane_vectors, 0, 0, grid_length_x, grid_length_y)

    # Plotting of the points
    for element in unused:
        plt.scatter(element[0], element[1], marker='s', s=130, c='#33638DFF')

    for element in used:
        plt.scatter(element[0], element[1], marker='D', s=130, c='#FDE725FF')

    # Plot of the crystal symmetry directions, tag must be activated.
    if symm_vec:
        import numpy.linalg as lin

        for element in symm_matrices:
            eig_val, eig_vec = lin.eig(element)

        for element in eig_vec:
            plt.quiver(
                *origin, element[0], element[1], alpha=1, color='#B8DE29FF', angles='xy', scale_units='xy', scale=1
            )

    # Plot of the Bravais lattice
    for element in plane_vectors:
        plt.quiver(*origin, element[0], element[1], color='#3CBB75FF', angles='xy', scale_units='xy', scale=1)

    legend_elements = [
        Line2D([0], [0], color='#33638DFF', lw=2, label='Unscanned Sites', marker='s'),
        Line2D([0], [0], color='#FDE725FF', lw=2, label='Scanned Sites', marker='D'),
        Line2D([0], [0], color='#3CBB75FF', lw=2, label='Bravais lattice'),
    ]
    plt.legend(handles=legend_elements, bbox_to_anchor=(0.75, -0.15))

    plt.title('Lattice plot and symmetry directions')
    plt.ylabel('y direction')
    plt.xlabel('x direction')
    #plt.xticks(np.arange(-grid_length, grid_length, float(plane_vectors[0][0])))
    #plt.set_cmap(cmap)
    plt.grid(linestyle='--')
    plt.show()


##########################################################
# find linear combination coefficients


def find_linear_combination_coefficients(plane_vectors, vectors):
    from operator import itemgetter
    """This helper function takes the planar vectors and a list of vectors
       and return the coefficients in the base of the planar vectors

       inputs ::

       plane vectors: list : list of list of the form [[a_x, a_y], [b_x, b_y]]

       returns ::

       indices : list : list of list of the form [[int_1, int_1]...[int_n, int_n]]
                        the integers refers to how many times that specific vectors is
                        present in the linear combination r = int_x * a + int_2 * b

       """

    # Formulate the system of equations Ax = b
    A = np.vstack((plane_vectors[0], plane_vectors[1])).T
    # inverse of A matrix
    Ainv = np.matrix(A)**(-1)
    indices = []
    # loop over flattened list
    for vec in np.array(vectors).reshape(-1, 2):
        # get indices from x = A^{-1}.b
        index = np.array((Ainv * np.matrix(vec).transpose()).transpose()).reshape(2)
        indices.append(index)
    # make sure to have integer values
    indices = np.round(np.array(indices), 0)

    isort = indices[:, 0] + indices[:, 1] / (10 * max(indices[:, 0]))
    indices = indices[isort.argsort()]

    return indices
